import re
from collections.abc import Iterator, Iterable
from typing import NamedTuple


class Token(NamedTuple):
    """A macro or text section found in the template by the tokenizer"""

    type: str
    """ The token type (error, text, embedded_macro, line_block_macro) """
    content: str
    """ The macro code of the macro section, resp. the text of the text section"""
    section_start_pos: int
    """ Position of first character (might be whitespace) of the macro section """
    start_marker_pos: int
    """ Position of first character of the introducing macro marker """
    content_pos: int
    """ Position of the first character of the token content in the template """


TokenStream = Iterator[Token]


class Tokenizer:
    """Text tokenizer for macro expansion. It extracts macro sections and
    remaining text sections and recognizes if a macro section is started but not ended.
    It can be configured by replacing the default regular expression patterns by
    application specific ones.

    :param macro_marker: If a string literal of the allowed forms or a comment starts
        with a string matching the *macro_marker*, it is treated as macro call.
    :param string_literal_start: Pattern for the start of a macro string literal
    :param string_literal_start_group: Name of the regular expression pattern group
        that encapsulates the concrete string used in the input to start the
        macro string literal. See *string_literal_end*.
    :param string_literal_end: End of a macro string literal.
        If *string_literal_start* allows several options to start a string literal,
        *string_literal_end* can refer to *string_literal_start_group*
        to express, that the literal needs to end like it started.
    :param line_comment_start: Pattern for the start of a macro string in a comment.
    """

    def __init__(
        self,
        macro_marker: str = r"\$\$",
        string_literal_start: str = r"(''')|'" + r'|(""")|"',
        string_literal_start_group: str = "pm4p_quotes",
        string_literal_end: str = r"(?P=pm4p_quotes)",
        line_comment_start: str = r"#(( |\t)*)",
    ) -> None:
        # Prefixes used for distinguishing the named match groups of the tokenizer
        # from others that might be used in the patterns chosen by the application
        self._token_group_prefix = token_group_prefix = "pm4p_grp_"
        self._token_pos_group_prefix = token_pos_group_prefix = "pm4p_pos_grp_"
        token_other_group_prefix = "pm4p_other_grp_"

        # Helper functions and literals for regular expressions
        def re_in_brackets(pattern: str) -> str:
            return r"(" + pattern + r")"

        def re_named_group(name: str, content_pattern: str) -> str:
            return re_in_brackets(r"?P<" + name + r">" + content_pattern)

        def re_or_bracketed_elements(patterns: Iterable[str]) -> str:
            """Regular expression or'ing some patterns. The patterns are put into
            brackets to avoid associativity problems, the result is not."""
            return re_in_brackets(r")|(".join(patterns))

        def re_not_ahead(pattern: str) -> str:
            return re_in_brackets(r"?!" + pattern)

        def re_if(id_or_name: str, yes_pattern: str, no_pattern: str) -> str:
            return re_in_brackets(
                r"?" + re_in_brackets(id_or_name) + yes_pattern + r"|" + no_pattern
            )

        optional_whitespace = r"(\s*)"
        spaces_or_tabs = r"([ \t]*)"
        character_or_newline = re_in_brackets(r".|\n")
        anything = character_or_newline + r"+"
        anything_non_greedy = character_or_newline + r"+?"
        eol = r"$"
        sol = r"^"
        end_of_line_optionally_nl = re_in_brackets(r"$\n?")

        # Regular expression for a macro call
        # (Starts with quotes or like a comment, continues with a macro marker,
        # continues with macro text, and ends with matching closing quotes,
        # if it has started with brackets, or the end of the line.)
        def re_macro_instance(
            group_suffix: str,
        ) -> str:
            # Adapt the named group names in order to make them usable multiple times
            # within the same regular expression
            local_start_marker_group = token_pos_group_prefix + group_suffix
            local_string_literal_start_group = string_literal_start_group + group_suffix
            local_string_literal_end = string_literal_end.replace(
                string_literal_start_group, local_string_literal_start_group
            )

            # starting quotes or hash. Quotes have their own group to describe
            # matching closing quotes later on. All together has a group to know
            # the starting position of non-whitespace part of the macro section.
            local_start_marker = (
                re_named_group(
                    local_start_marker_group,
                    re_or_bracketed_elements(
                        [
                            re_named_group(
                                local_string_literal_start_group, string_literal_start
                            ),
                            line_comment_start,
                        ]
                    ),
                )
                + macro_marker
                + optional_whitespace
            )
            local_end_marker = re_if(
                local_string_literal_start_group,
                optional_whitespace + macro_marker + local_string_literal_end,
                eol,
            )
            return (
                local_start_marker
                + re_named_group(token_group_prefix + group_suffix, anything_non_greedy)
                + local_end_marker
            )

        def re_macro_line_block_instance(
            group_suffix: str,
        ) -> str:
            # 'line_block_macro'
            return re_in_brackets(
                sol
                + spaces_or_tabs
                + re_macro_instance(group_suffix)
                + spaces_or_tabs
                + end_of_line_optionally_nl
            )

        # Regular expression for a text block
        # ("as long as it does not look like a line block macro or the start
        # of a non-line-block macro section". The first is necessary, because
        # a match of a line block macro can also fail by the text after
        # the macro section. And the second case need to be limited to the start
        # of the section in order to find syntax errors of missing section endings.)
        re_text = re_named_group(
            token_group_prefix + "text",
            re_in_brackets(
                re_not_ahead(
                    re_or_bracketed_elements(
                        [
                            re_macro_line_block_instance(
                                token_other_group_prefix + "ahead_line_block"
                            ),
                            re_in_brackets(string_literal_start) + macro_marker,
                            re_in_brackets(line_comment_start) + macro_marker,
                        ]
                    )
                )
                + character_or_newline
            )
            + r"+",
        )

        # Regular expression that matches anything else and reports an error
        re_error = re_named_group(token_group_prefix + "error", anything)

        # A token is a macro call, or (otherwise) a block that does not look
        # like the start of a macro call, or (otherwise) a syntax error
        # (text, that starts like a macro call, but does not fully match the
        # syntax of a macro call)
        re_tokens = re_or_bracketed_elements(
            [
                re_macro_line_block_instance("line_block_macro"),
                re_macro_instance("embedded_macro"),
                re_text,
                re_error,
            ]
        )
        # import sys
        # print(re_tokenizer, file=stderr)
        # raise RuntimeError("tmp")
        self._tokenizer_pattern = re.compile(re_tokens, re.MULTILINE)

    def tokenize(self, text: str) -> TokenStream:
        """Iterate and unpack oll tokens in *text*."""

        # A token is a match of a named group with a name started by
        # *token_group_prefix*.
        tokenizer_pattern = self._tokenizer_pattern
        token_group_prefix = self._token_group_prefix
        for match in re.finditer(tokenizer_pattern, text):
            # Get match data
            group_dict, pos = match.groupdict(), match.start()
            if not group_dict:  # pragma: no cover
                raise RuntimeError(f"Internal error: No match result at pos {pos}")

            # Groups that match and are relevant, because they stem from the tokenizer
            found_tokens = {
                name: token_content
                for (name, token_content) in group_dict.items()
                if token_content is not None
                and name.removeprefix(token_group_prefix) != name
            }
            if not found_tokens:  # pragma: no cover
                raise RuntimeError(f"Internal error: No match at pos {pos}")
            # Get group with the syntax case we are in, and the code content
            if len(found_tokens) > 1:  # pragma: no cover
                raise RuntimeError(
                    f"Internal error: Several matches at pos {pos}" f": {found_tokens}"
                )
            token_group_name, token_content = found_tokens.popitem()
            token_type = token_group_name.removeprefix(token_group_prefix)
            token_content_pos = match.start(token_group_name)

            # start position of section
            section_start_pos = match.start()

            # In case of macro: position of start marker
            start_marker_pos = (
                match.start(self._token_pos_group_prefix + token_type)
                if token_type in ["macro", "line_block_macro"]
                else section_start_pos
            )

            yield Token(
                token_type,
                token_content,
                section_start_pos,
                start_marker_pos,
                token_content_pos,
            )



import unittest
import pymacros4py
import pathlib
import tempfile
import os
import re
import contextlib
import io
import subprocess


class ReadmeExamplesTest(unittest.TestCase):
    def test_expansion_result(self) -> None:
        """
        Test that all README examples have the specified expansion script and result
        """
        for template_path in pathlib.Path("tests/data/").glob("doc_*.tpl.py"):
            with self.subTest(template=str(template_path), mode="script"):
                result_file_path = template_path.with_suffix("").with_suffix(
                    ".script.py"
                )
                pp = pymacros4py.PreProcessor()
                result = pp.template_script(str(template_path))
                # Hide internals from the examples for the README.rst
                result = re.sub(r"^ *_macro_.*$\n", "", result, flags=re.MULTILINE)
                if result_file_path.exists():
                    prev_result = pymacros4py._files.read_file(str(result_file_path))
                    self.assertMultiLineEqual(result, prev_result)
                else:
                    with open(result_file_path, "w", encoding="utf-8") as f_out:
                        f_out.write(result)

            with self.subTest(template=str(template_path), mode="expand"):
                result_file_path = template_path.with_suffix("").with_suffix(".py")
                pp = pymacros4py.PreProcessor()
                if result_file_path.exists():
                    result = pp.expand_file(str(template_path))
                    prev_result = pymacros4py._files.read_file(str(result_file_path))
                    self.assertMultiLineEqual(result, prev_result)
                else:
                    pp.expand_file_to_file(str(template_path), str(result_file_path))


class ReadmeTest(unittest.TestCase):
    overwrite = False

    def test(self) -> None:
        """
        Test that README.rst is up-to-date. Create, if missing.
        """

        pp = pymacros4py.PreProcessor()
        templates_root = pathlib.Path("tpl")
        for template in templates_root.glob("**/*"):
            if not template.is_file():
                continue
            goal = template.relative_to(templates_root)
            if goal.exists() and not self.overwrite:
                result = pp.expand_file_to_file(
                    str(template), str(goal), diffs_to_result_file=True
                )
                if result:
                    print(result, flush=True)
                    self.fail(str(goal) + " not up to date. See differences above.")
            else:
                pp.expand_file_to_file(str(template), str(goal))


class TestForMentionedButNotDocumentedBehaviour(unittest.TestCase):
    def test_insert_from_with_caching(self) -> None:
        """The README says, that repeated calls of *insert_from*
        with the same file and without parameter *global* leads to re-use
        of already expanded template data. The template generates a random
        number for each time a template is expanded. They equal, if only
        one expansion happened."""
        template_path = "tests/data/testcase_insert_from_with_caching.tpl.py"
        pp = pymacros4py.PreProcessor()
        result = pp.expand_file(template_path).splitlines()
        self.assertEqual(result[0], result[1])

    def test_insert_from_without_caching(self) -> None:
        """The README says, that repeated calls of *insert_from*
        with the same file and with non-None parameter *global* disables
        re-use of already expanded template data. For more see
        test_insert_from_with_caching."""
        template_path = "tests/data/testcase_insert_from_without_caching.tpl.py"
        pp = pymacros4py.PreProcessor()
        result = pp.expand_file(template_path).splitlines()
        self.assertNotEqual(result[0], result[1])

    def test_no_re_import(self) -> None:
        """The README says, that repeated calls of *import_from*
        with the same file is blocked. The template import another template
        that sets a variable to a new random value each time the import
        happens. Then, we can see, whether the second import is prevented."""
        template_path = "tests/data/testcase_import_from_a_second_time.tpl.py"
        pp = pymacros4py.PreProcessor()
        result = pp.expand_file(template_path).splitlines()
        self.assertEqual(result[0], result[1])

    def test_writing_result_files(self) -> None:
        """The README says, that expand_file_to_file can write its results
        to a file. All other tests just verify results, if the output already
        exists in the target file. Thus, this test here tests writing."""
        template_path = "tests/data/doc_templ_and_templ_exp.tpl.py"
        tmp_file, tmp_file_path = tempfile.mkstemp(text=True)
        try:
            pp = pymacros4py.PreProcessor()
            result = pp.expand_file(template_path)
            pp.expand_file_to_file(template_path, tmp_file_path)
            result_in_file = pymacros4py._files.read_file(tmp_file_path)
            self.assertEqual(result, result_in_file)
        finally:
            os.close(tmp_file)

    def test_trace_parsing_with_exception(self) -> None:
        """The README describes exception handling and tracing. The
        indirect expectation is, that an exception does not prevent the
        tracing from working. There is no explicit example in the README,
        so, we test this here, for tracing the parsing."""
        template_path = "tests/data/testcase_tracing_parsing_and_exception.tpl.py"
        pp = pymacros4py.PreProcessor()
        with contextlib.redirect_stdout(io.StringIO()) as f:
            self.assertRaisesRegex(
                RuntimeError,
                r".*Syntax error in macro section, macro started but not ended.*",
                pp.expand_file,
                template_path,
                trace_parsing=True,
            )
        s = f.getvalue()
        self.assertRegex(
            s, r"(?s)" r".*line 1.*text.*x = " r".*line 1.*error.*insert\(v\).*"
        )

    def test_trace_evaluation_with_exception(self) -> None:
        """The README describes exception handling and tracing. The
        indirect expectation is, that an exception does not prevent the
        tracing from working. There is no explicit example in the README,
        so, we test this here, for tracing the evaluation."""
        template_path = "tests/data/testcase_tracing_evaluation_and_exception.tpl.py"
        pp = pymacros4py.PreProcessor()
        with contextlib.redirect_stdout(io.StringIO()) as f:
            # The expected behaviour depends on the availability of method
            # *add_note* for exceptions.
            if hasattr(RuntimeError(), "add_note"):
                self.assertRaisesRegex(
                    NameError,
                    r"name 'unknown_variable' is not defined",
                    pp.expand_file,
                    template_path,
                    trace_evaluation=True,
                )
            else:
                self.assertRaisesRegex(
                    RuntimeError,
                    r"Error occurred when expanding the following template:*",
                    pp.expand_file,
                    template_path,
                    trace_evaluation=True,
                )
        s = f.getvalue()
        self.assertRegex(
            s,
            r"(?s)"
            r".*line 1.*line_block_macro.*insert"
            r".*line 2.*line_block_macro.*insert.*",
        )

    def test_run_process_with_file_with_exception(self) -> None:
        """The README describes the behaviour when a
        *subprocess.CalledProcessError* occurs, but gives no doctest."""
        with contextlib.redirect_stdout(io.StringIO()) as f:
            self.assertRaisesRegex(
                subprocess.CalledProcessError,
                r"returned non-zero exit status 1",
                pymacros4py.run_process_with_file,
                ["python", "-m", "non_existing_module"],
                "some_file_path",
            )
        s = f.getvalue()
        self.assertRegex(
            s,
            r"(?s)"
            r".*The called process raised an exception"
            r".*some_file_path"
            r".*No module named non_existing_module.*",
        )

    def test_import_from_with_error(self) -> None:
        """The README describes importing from files. The
        indirect expectation is, that an exception also informs about the
        template that is expanded while the exception is raised.
        here is no explicit example in the README,
        so, we test this here."""
        template_path = "tests/data/testcase_import_from_with_error.tpl.py"
        pp = pymacros4py.PreProcessor()
        with contextlib.redirect_stdout(io.StringIO()) as f:
            self.assertRaisesRegex(
                RuntimeError,
                r".*Exception when importing.*error_unexpected_end.tpl.py.*",
                pp.expand_file,
                template_path,
                trace_parsing=True,
            )
        s = f.getvalue()
        self.assertRegex(
            s, r"(?s).*testcase_import_from_with_error.tpl.py.*import_from.*"
        )


if __name__ == "__main__":
    # This code allows to start the input/output test cases manually, without
    # all other tests triggered by the test procedure for the whole package.
    # Note: If this module is run on itself, the current directory needs to
    # be the base directory of the package.
    ReadmeExamplesTest().test_expansion_result()



from typing import Optional
import tempfile
import os
import sys
import pathlib
from dataclasses import dataclass

from ._files import read_file
from ._tokenizer import Tokenizer
from ._global_evaluation_context import GlobalEvaluationContext
from ._template_script import TemplateScript


@dataclass
class Macro:
    outer_macro: Optional["Macro"]
    indentation: str
    is_embedded: bool
    content_line: str
    output: list[str]


def _separate_indentation_and_content_optionally_nl(s: str) -> tuple[str, str]:
    """Get the whitespace characters, that the *text* starts with, and the rest of it

    :param s: A string without newline characters except, maybe, the last character
    """

    # Separate the line and a possible trailing single newline
    if s and s[-1] == "\n":
        s, newline = s[:-1], s[-1]
    else:
        s, newline = s, ""

    # Strip any whitespace (s has no NL anymore) starting from the left
    s_stripped = s.lstrip()
    # Extract exactly the stripped whitespaces
    whitespace_len = len(s) - len(s_stripped)

    # Return the leading whitespace if s, and the following content (optionally with
    # the newline, we put aside at the beginning
    return s[0:whitespace_len], s_stripped + newline


def evaluate_template_script(
    template_script: TemplateScript,
    tokenizer: Tokenizer,
    global_evaluation_context: GlobalEvaluationContext,
    already_imported_files: set[str],
    globals_dict: Optional[dict] = None,
) -> str:
    """
    Run the *template_script* and return the results. For recursive
    template expansion, use *tokenizer* in the *global_evaluation_context*
    with the *globals_dict*.

    :param template_script: The script to evaluate.
    :param tokenizer: The tokenizer for recursive template expansions.
    :param global_evaluation_context: Global context information for all template
        expansions happening under a single PreProcessor.
    :param already_imported_files: Template files that have already been "imported"
        along the current (possibly recursive) expansion based on the globals_dict.
        Used by method *import_from* to avoid repeated imports of content
        from the same file.
    :param globals_dict: Like parameter *globals* of the Python function
        *exec*. Note, that in order to use it, a value for key *pp* will be
        set.
    """

    # The following two variables are also used within the functions, that are provided
    # to the expansion script as part of the content of its global namespace:
    # List of all generated fragments of output
    output: list[str] = []
    # Subsequent output does not stem from a macro
    macro: Optional[Macro] = None

    # The following functions are the respective functions that the template expansion
    # code can use.
    # They are defined as closures, in order to be able to access the current
    # evaluation context, the variables defined above, without needing an access path
    # seen in the template script.

    def _macro_starts(indentation: str, embedded: bool, content_line: str) -> None:
        """The *macro_starts_ function for macro code.

        Store the information, that subsequent *insert* calls stem from
        the current macro, the standard indentation for expansion results
        inserted by these calls, and whether the current macro is an embedded
        macro.

        The first line inserted by an embedded macro remains without indentation.
        """
        nonlocal macro
        macro = Macro(macro, indentation, embedded, content_line, list[str]())

    def _macro_ends(content_line: str) -> None:
        """The *macro_ends* function for macro code.

        Store the information, that subsequent *insert* calls do not stem
        from the current macro anymore.
        """
        nonlocal macro, output

        if macro is None:  # pragma: no cover
            raise RuntimeError(
                f"{content_line}:\n"
                f"Internal error: macro_end without macro_start in the "
                f"template expansion script for the above given template line"
            )

        current_macro = macro
        macro = current_macro.outer_macro
        current_output = macro.output if macro else output

        # Handle macro output (expansion result) line by line
        lines = "".join(current_macro.output).splitlines(keepends=True)

        if not lines:
            # Not a single line to process: finished
            return

        # First line
        line = lines.pop(0)
        (
            base_indentation,
            line_content,
        ) = _separate_indentation_and_content_optionally_nl(line)
        if current_macro.is_embedded:
            # Use unindented result content
            current_output.append(line_content)
        else:
            # Indent as defined for the macro
            current_output.append(current_macro.indentation + line_content)

        # Each subsequent line:
        for line in lines:
            (
                line_indentation,
                line_content,
            ) = _separate_indentation_and_content_optionally_nl(line)
            if len(line_indentation) == 0 and len(base_indentation) > 0:
                # Zero indentation in a context with none-zero base indentation:
                # Just take the line as it is
                current_output.append(line)
            elif line_indentation[0 : len(base_indentation)] == base_indentation:
                # Line indentation string is extension of base indentation
                # string: Compute relative indentation (including the content), and
                # prepend macro indentation to it
                current_output.append(
                    current_macro.indentation + line[len(base_indentation) :]
                )
            else:
                # If the line is indented, indentation (as a string)
                # need to start with exactly the base indentation
                raise RuntimeError(
                    f"{current_macro.content_line}:\n"
                    f"Output syntax error: indentation of the following line of the "
                    f"results of the template script from the above given template "
                    f"line is not an extension of the base indentation of these "
                    f"results:\n"
                    f">{line.rstrip()}<\n"
                    f"(Start of line shown enclosed by characters '>' and '<')"
                )

    def insert(*vargs: object) -> None:
        """The *insert function* for macro code.

        Concatenate string representations (result from *str()*)
        of the arguments. Indent the lines with the current indentation,
        except for the first line produced by an embedded macro.
        Insert result to the result of the template expansion.
        """
        nonlocal macro, output
        lines_str = "".join((str(a) for a in vargs))

        if lines_str == "":
            # Nothing to insert into expansion output
            return

        current_output = macro.output if macro else output
        # If we have been called from within macro code (this includes the case that
        # the macro code calls a function that contains a text block that calls
        # us): store content for bulk processing at the end of the macro
        # If we have been called from a text block (and it is not part
        # of a function definition within a macro): Output it as-is
        # Both is implemented the same way.
        current_output.append(lines_str)

    def insert_content(file: str) -> None:
        """
        Insert the content of the file to the expansion results of the current
        template expansion. Do not perform a template expansion.

        :param file: File to process.
        """
        insert(read_file(file))

    def insert_from(
        template_file: str,
        globals_dict: Optional[dict] = None,
        trace_parsing: bool = False,
        trace_evaluation: bool = False,
    ) -> None:
        """
        Perform a macro expansion on the content of a file in the scope
        of a new Python interpreter.
        Use *globals* to initialize the interpreter like in a call of *eval()*.

        Insert the output to the expansion results of the current template
        expansion.

        When called a second time with an identical argument for *template_file*,
        and *globals* is *None* in both calls, re-use the output of the previous run.

        (If *globals* is not *None*, and you like to re-use results in cases of
        equivalent content of *globals*, this has to be implemented manually.)

        :param template_file: Template to expand. If the first part of the path
          (see *pathlib.PurePath.parts*) is '$$', this part is removed and the
          subsequent parts are interpreted relative to the directory of the
          currently expanded template.
        :param globals_dict: Like parameter *globals* of the Python function *exec*.
          Note, that in order to use it, a value for key *macro* will be set.
        :param trace_parsing: Print parsing log to stderr.
        :param trace_evaluation: Print evaluation log to stderr.
        """
        parts = pathlib.PurePath(template_file).parts
        if parts and parts[0] == "$$":
            template_dir = pathlib.PurePath(template_script.file_name).parent
            template_file = str(pathlib.PurePath(template_dir, *parts[1:]))
        template = read_file(template_file)
        template_script_to_insert_from = TemplateScript(
            template_file, template, tokenizer, trace_parsing, trace_evaluation
        )

        if globals_dict:
            # Here, we cannot cache, because we cannot recognize identical
            # content of the globals_dict
            result = evaluate_template_script(
                template_script_to_insert_from,
                tokenizer,
                global_evaluation_context,
                already_imported_files,
                globals_dict,
            )
        else:
            already_inserted = global_evaluation_context.already_inserted_content
            template_file_resolved = str(
                pathlib.Path(template_file).resolve(strict=True)
            )
            if template_file_resolved in already_inserted:
                result = already_inserted[template_file_resolved]
            else:
                result = evaluate_template_script(
                    template_script_to_insert_from,
                    tokenizer,
                    global_evaluation_context,
                    set[str](),  # no imports so far, due to new evaluation context
                    None,  # empty globals dict -> new evaluation context
                )
                already_inserted[template_file_resolved] = result
        insert(result)

    def import_from(
        template_file: str,
        trace_parsing: bool = False,
        trace_evaluation: bool = False,
    ) -> None:
        """
        Perform a macro expansion of *file* in the scope of the current interpreter.
        Attributes that have already been set can be used by macro code in
        *template_file*,
        and attributes set by such code can be used in macro code following
        the call.

        Discard the output of the expansion run.

        When called a second time with an identical argument for *template_file*,
        ignore the call.

        :param template_file: Template to expand. If the first part of the path
          (see *pathlib.PurePath.parts*) is '$$', this part is removed and the
          subsequent parts are interpreted relative to the directory of the
          currently expanded template.
        :param trace_parsing: Print parsing log to stderr.
        :param trace_evaluation: Print evaluation log to stderr.
        """

        parts = pathlib.PurePath(template_file).parts
        if parts and parts[0] == "$$":
            template_dir = pathlib.PurePath(template_script.file_name).parent
            template_file = str(pathlib.PurePath(template_dir, *parts[1:]))
        template_file_resolved = str(pathlib.Path(template_file).resolve(strict=True))
        if template_file_resolved in already_imported_files:
            return
        try:
            template = read_file(template_file)
            template_script_to_import_from = TemplateScript(
                template_file, template, tokenizer, trace_parsing, trace_evaluation
            )
            _ = evaluate_template_script(
                template_script=template_script_to_import_from,
                tokenizer=tokenizer,
                global_evaluation_context=global_evaluation_context,
                already_imported_files=already_imported_files,
                globals_dict=globals_dict,
            )
            already_imported_files.add(template_file_resolved)
        except Exception as e:
            raise RuntimeError(
                f'Exception when importing from file "{template_file}"'
            ) from e

    # Globals dict for the execution of the template script:
    # Make the current object accessible from within the template script code, and
    # also the functions that are meant to be called from within this code.
    # globals_dict["pp"] = global_evaluation_context
    globals_to_set = {
        "insert": insert,
        "insert_content": insert_content,
        "insert_from": insert_from,
        "import_from": import_from,
        "_macro_starts": _macro_starts,
        "_macro_ends": _macro_ends,
        "stderr": sys.stderr,
    }

    # Make these assignments, and prepare for undoing them later, is necessary
    if globals_dict is None:
        globals_dict = globals_to_set
        globals_backup = dict()
    else:
        globals_backup = {
            key: globals_dict[key]
            for key in globals_to_set.keys()
            if key in globals_dict
        }
        globals_dict.update(globals_to_set)

    # We create an empty file, that will not (!) be removed when
    # an exception happens. We use its path as virtual name
    # when executing a script that is stored in a string.
    # When the exec raises an exception, we store the template
    # that we have already executed to the file, so that the user
    # can "debug" the file that the exception mentions.
    # We need individual files since macro expansion can be nested,
    # so that we get different temporary files for each nesting level.
    tmp_file, tmp_file_path = tempfile.mkstemp(
        suffix=".py",
        prefix=f"template_script_{next(global_evaluation_context.tmp_file_numbers)}",
        dir=None,
        text=True,
    )

    # Execute the template script and return what it reports using *insert*
    template_script_code = str(template_script)
    try:
        ast_object = compile(
            template_script_code,
            tmp_file_path,
            mode="exec",
            # flags=0, dont_inherit=False, optimize=- 1
        )
        exec(ast_object, globals_dict)

        # Exec raised no exception, so we do not need the temporary file.
        os.close(tmp_file)
        os.remove(tmp_file_path)

        return "".join(output)

    except Exception as exc:
        # save the template to the file and close it
        try:
            template_as_bytes = template_script_code.encode(
                encoding="utf-8", errors="strict"
            )
            os.write(tmp_file, template_as_bytes)
        finally:
            os.close(tmp_file)
        note = (
            "Error occurred when executing template script.\n"
            f' File "{tmp_file_path}"'
        )
        # Depending on the used Python version, one of the following will happen.
        if hasattr(exc, "add_note"):  # pragma: no cover
            exc.add_note(note)
            raise
        raise RuntimeError(note) from exc  # pragma: no cover
    finally:
        # If a globals_dict has been given to us, undo changes we have done there
        globals_dict.update(globals_backup)



import itertools


class GlobalEvaluationContext:
    """
    Global context information for all template expansions happening under a single
    PreProcessor.
    """

    def __init__(self) -> None:
        self.already_inserted_content = dict[str, str]()
        # A cache of the expansion results of all template files that have already
        # been "inserted". Used by method *insert_from* of the evaluator to avoid
        # unnecessary repeated expansions.

        self.tmp_file_numbers = itertools.count()
        # Global counter for generated temporary files. During recursive evaluation,
        # all files 'on the stack' exist in parallel, so they are numbered.
        # Currently, an exception stops the evaluation and only the files on the
        # stack remains (for debugging), but for the future, it might be possible
        # to continue expansion with other files, and for this, be use a global
        # counter here, and not one just for files on the local evaluation stack.



import os
import tempfile
import subprocess
from typing import TypeAlias, Optional
from dataclasses import dataclass, asdict


FileName: TypeAlias = str | bytes | os.PathLike


@dataclass
class FileOptions:
    """Options as they can be used for opening a file. See Python function *open*."""

    encoding: Optional[str] = None
    errors: Optional[str] = None
    newline: Optional[str] = None


file_options = FileOptions(None, None, None)
""" Options used for reading and writing files. See Python function *open*. """


def read_file(in_file_name: FileName, finally_remove: bool = False) -> str:
    """Load text from *in_file_name* using the chosen *file_options*.
    If option *remove_on_error* is set to True and an exception occurs,
    remove the file."""
    try:
        with open(in_file_name, **asdict(file_options)) as f_in:
            s_in = f_in.read()
    finally:
        if finally_remove:
            os.remove(in_file_name)
    return s_in


def write_file(out_file_name: FileName, content: str) -> None:
    """Write text to *out_file_name* using the chosen *file_options*."""
    with open(out_file_name, "w", **asdict(file_options)) as f_out:
        f_out.write(content)


def write_to_tempfile(content: str) -> str:
    """Write text to a temporary file using the chosen *file_options*
    and return the path of the file as str."""
    tmp_file, tmp_file_name = tempfile.mkstemp(text=True)
    os.close(tmp_file)

    try:
        write_file(tmp_file_name, content)
    except Exception as e:  # pragma: no cover
        os.remove(tmp_file_name)
        raise e

    return str(tmp_file_name)


def run_process_with_file(args: list[str], path_str: str) -> None:
    """Run the command described by *args* (compare subprocess.run)
    using *subprocess.run* and capture both stdout and stderr. If a
    *subprocess.CalledProcessError is raised, print the captured output
    and print an error messages that hints to file *path_str*.

    Example:
    pymacros4py.run_on_tempfile(["black", "path_to_file"], "path_to_file")
    """
    try:
        subprocess.run(
            args,
            # capture_output=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            encoding="utf8",
            check=True,
        )
    except subprocess.CalledProcessError as e:
        print(
            "\n"
            "The called process raised an exception.\n"
            "The file it operated on starts here:\n"
            f'  File "{path_str}", line 0'
        )
        print("The process returned as output:")
        print(str(e.output))
        raise



from ._tokenizer import Tokenizer
from ._pre_processor import PreProcessor
from ._files import (
    file_options,
    read_file,
    write_file,
    write_to_tempfile,
    run_process_with_file,
)

__all__ = (
    # ._tokenizer
    "Tokenizer",
    # ._pre_precessor
    "PreProcessor",
    # ._files
    "file_options",
    "read_file",
    "write_file",
    "write_to_tempfile",
    "run_process_with_file",
)



import itertools
import re

from ._tokenizer import Tokenizer


class TemplateScriptIndentation:
    """
    Current indention state of a stream of created code. Used during the creation of a
    template script."""

    def __init__(self, steps: int) -> None:
        self._indentation_level = 0
        self._indentation_steps = steps

    def indent(self) -> None:
        self._indentation_level += 1

    def dedent(self, content_line: str, content: str) -> None:
        self._indentation_level -= 1
        if self._indentation_level < 0:
            raise RuntimeError(
                f"--- {content_line}: "
                "Nesting error in compound statements "
                "with suites spanning several sections, "
                f"in macro section:\n{content}"
            )

    def __str__(self) -> str:
        return " " * self._indentation_level * self._indentation_steps

    def __bool__(self) -> bool:
        return self._indentation_level != 0


def _extract_prefix(template: str, pos: int) -> str:
    """Find the start of the line in *template* where *pos* is.
    Return the characters from the start of the line till the one
    before *pos*.

    :param template: A template. Needs to follow Unix-style end of line convention.
    :param pos: A position (number of characters) in the template
    """
    # Find position of previous newline character and skip it.
    # If we do not find one, we are at the first line of the text, get
    # position -1, and the '+1' brings us to the first character of the file.
    nl_pos = template.rfind("\n", 0, pos) + 1
    # characters from line start (inclusively) to pos (exclusively)
    return template[nl_pos:pos]


def _line_no(template: str, pos: int) -> int:
    """Return the line in *template* where *pos* is, counted from 1"""
    return template[:pos].count("\n") + 1


# def _col_no(template: str, pos: int) -> int:
#     """ Return the column in *template* where *pos* is, counted from 1 """
#     return pos - template[:pos].rfind("\n") + 1


def _separate_indentation_and_content(text: str) -> tuple[str, str]:
    """Get the whitespace characters, that the *text* starts with, and the rest of it

    :param text: A string without newline characters
    """
    # strip any whitespace starting from the left
    text_stripped = text.lstrip()
    # Extract exactly the stripped whitespaces
    whitespace_len = len(text) - len(text_stripped)
    return text[0:whitespace_len], text_stripped


class TemplateScript:
    """
    Template script for *template*, created using *tokenizer*.
    Its string representation (as returned by *str(template_script)*) is the
    generated script code.

    :param file_name: Name of the file of the template. Used for log messages.
    :param template: The template to expand. Needs to follow Unix-style
       end of line convention.
    :param tokenizer: Tokenizer to use.
    :param trace_parsing: Print parsing log.
    :param trace_evaluation: Print evaluation log.
    """

    # Given the current functionality and use cases, the class could be replaced by a
    # function that directly returns the generated script code. The form of a class has
    # been chosen to make it possible to add further attributes in the future.

    def __init__(
        self,
        file_name: str,
        template: str,
        tokenizer: Tokenizer,
        trace_parsing: bool = False,
        trace_evaluation: bool = False,
    ) -> None:
        self.file_name = file_name

        # Token texts that have already been regarded in the generation of the
        # token expansion code. Used to give the template script access to them.
        token_strings = list[str]()

        # Initialize handling of indentation in template script
        script_indentation = TemplateScriptIndentation(4)

        # List of the strings produced by the expansion of the found tokens
        template_script_strings = []

        # Parse tokens and generate template expansion code, string by string
        for token_no, token in zip(itertools.count(), tokenizer.tokenize(template)):
            (
                token_type,
                content,
                section_start_pos,
                start_marker_pos,
                content_pos,
            ) = token

            content_line = f'File "{file_name}", line {_line_no(template, content_pos)}'

            if trace_parsing:
                print(f"--- {content_line}: {token_type}:\n>{content}<\n\n", flush=True)

            token_strings.append(content)

            if trace_evaluation:
                template_script_strings.append(
                    f"{str(script_indentation)}"
                    f"print('''{repr(content_line)}: {token_type}\n"
                    f">{token_strings[token_no]}<\n\n''', flush=True)\n"
                )

            if token_type == "error":
                raise RuntimeError(
                    f"--- {content_line}: "
                    "Syntax error in macro section, macro started but not ended:\n"
                    f"{content}"
                )

            elif token_type == "text":
                s = f"{str(script_indentation)}insert({repr(content)})\n"
                template_script_strings.append(s)

            elif token_type in ["embedded_macro", "line_block_macro"]:
                # Section indentation
                start_marker_prefix = _extract_prefix(template, start_marker_pos)
                start_marker_indentation = re.sub(r"\S", " ", start_marker_prefix)

                # Base indentation
                code_start_prefix = _extract_prefix(template, content_pos)
                base_indentation = re.sub(r"\S", " ", code_start_prefix)

                # If the last character of macro code is a colon,
                # the macro code is a compound statement with a multi-section suite.
                # (We can check this like this, because whitespace has already been
                # stripped on the right by the tokenizer)
                multi_section_suite_starts = content and content[-1] == ":"

                # Case that content is":end":
                # (We can check this like this, because other
                # content is not allowed in macro code together with ":end")
                multi_section_suite_ends = content.strip()[0:4] == ":end"

                if not multi_section_suite_starts and not multi_section_suite_ends:
                    # Inform the global_evaluation_context of the template script that a
                    # new macro starts here, what indention its output need to have, and
                    # whether it is an embedded macro.
                    template_script_strings.append(
                        str(script_indentation)
                        + "_macro_starts("
                        + "indentation="
                        + repr(start_marker_indentation)
                        + ", "
                        + "embedded="
                        + str(token_type == "embedded_macro")
                        + ", "
                        + "content_line="
                        + repr(content_line)
                        + ")\n"
                    )

                if multi_section_suite_ends:
                    # Handle suite (of a compound statement) ending in this macro
                    script_indentation.dedent(content_line, content)
                    # Note: A macro that ends a suite (of a compound statement)
                    # has no _macro_end (and no _macro_start).
                    continue

                # # Detect and expand shorthand notation "? something" for
                # # "insert(something)"
                # # (This functionality is currently not offered)
                # if content[0] == "?":
                #     content = "insert(" + content[1:].lstrip() + ")"

                # Case of a statement ending a suite and re-starting a new one
                if re.match(r"(elif|else|except|finally|case).*:", content):
                    script_indentation.dedent(content_line, content)

                # iterator of numbered lines
                lines = content.splitlines()
                numbers_and_lines = enumerate(lines)

                # Special-case first line of macro code: It already comes without
                # indentation
                number, line = next(numbers_and_lines)
                template_script_strings.append(str(script_indentation) + line + "\n")

                # All lines subsequent lines of output: De-indent line relative to
                # the base indentation, indent it for the results and insert it to
                # the results
                for no, line in numbers_and_lines:
                    line_indentation, _ = _separate_indentation_and_content(line)
                    if len(line_indentation) == 0 and len(base_indentation) > 0:
                        # Zero indentation in a context with none-zero base indentation:
                        # Just take the line as it is
                        template_script_strings.append(line + "\n")
                    elif (
                        line_indentation[0 : len(base_indentation)] == base_indentation
                    ):
                        # Line indentation string is extension of base indentation
                        # string: Compute relative indentation, and prepend
                        # script indentation
                        template_script_strings.append(
                            str(script_indentation)
                            + line[len(base_indentation) :]
                            + "\n"
                        )
                    else:
                        # If the macro code is indented, indentation (as a string)
                        # need to start with exactly the base indentation
                        raise RuntimeError(
                            f"{content_line}: "
                            f"Syntax error: indentation of line {no} of the "
                            f"macro code is not an extension of the base indentation."
                        )

                if multi_section_suite_starts:
                    # Handle compound statement with a suite beginning in this macro
                    # section and spanning subsequent sections
                    script_indentation.indent()

                if not multi_section_suite_starts:
                    # Inform the global_evaluation_context of the template script that
                    # the macro ends here
                    template_script_strings.append(
                        str(script_indentation) + f"_macro_ends({repr(content_line)})\n"
                    )

            else:  # pragma: no cover
                raise RuntimeError(
                    f"{content_line}:"
                    f"Internal error: Tokenizer returned unknown token "
                    f"type {token_type} at position {content_pos}"
                )

        # Check sanity of reached state
        if script_indentation:
            raise RuntimeError(
                "Syntax error: block nesting (indentation) not correct, "
                "is :end somewhere missing?"
            )

        # Concatenate the template strings to the template script
        self._template_script = "".join(template_script_strings)

    def __str__(self) -> str:
        """Return the code of the template script."""
        return self._template_script



if __name__ == "__main__":
    import doctest
    import unittest
    import coverage  # type: ignore

    # -- Start recording coverage
    cov = coverage.Coverage(source_pkgs=["pymacros4py"])
    cov.start()

    # -- Create empty TestSuite
    test_suite = unittest.TestSuite()
    verbosity = 1  # 1 normal, 2 for more details
    failfast = True  # True

    # -- Test modules: doc tests (currently: none)
    # for file in pathlib.Path("tests").glob("*.py"):
    #     file_name = file.name.removesuffix(".py")
    #     __import__(file_name)
    #     test_suite.addTests(doctest.DocTestSuite(file_name))

    # -- Test modules: unit tests classes
    new_suite = unittest.defaultTestLoader.discover("tests", pattern="test_unit*.py")
    test_suite.addTests(new_suite)

    # -- Package module source: doc tests (currently: none)
    # for file in pathlib.Path("src", "pymacros4py").glob("*.py"):
    #     file_name = file.name.removesuffix(".py")
    #     if file_name == "__init__":
    #         continue
    #     module = "." + file_name
    #     temp_module = importlib.import_module(module, "pymacros4py")
    #     test_suite.addTests(doctest.DocTestSuite(temp_module))

    # -- Separate documentation: doc tests (currently: none)
    # for file_path in pathlib.Path(".", "docs", "source").glob("*.rst"):
    #   test_suite.addTests(doctest.DocFileSuite(str(file_path), module_relative=False))

    # -- DocTests for README.rst
    test_runner = unittest.TextTestRunner(verbosity=verbosity, failfast=failfast)
    if test_runner.run(test_suite).wasSuccessful():
        # Create new empty TestSuite for doctests
        # (we do them separately to be sure, the README.rst is already created by the
        # unittests)
        test_suite = unittest.TestSuite()
        test_suite.addTests(doctest.DocFileSuite("README.rst", module_relative=False))
        unittest.TextTestRunner(verbosity=verbosity, failfast=failfast).run(test_suite)

    # -- Stop recording coverage, create HTML from results
    cov.stop()
    cov.save()
    cov.xml_report()
    cov.html_report()



import difflib
from typing import Optional

from ._tokenizer import Tokenizer
from ._template_script import TemplateScript
from ._files import read_file, write_file, FileName
from ._global_evaluation_context import GlobalEvaluationContext
from ._evaluator import evaluate_template_script


class PreProcessor:
    """Text pre-processor that separates macro code sections and
    text sections, generates a template script for them, executes
    this script, and uses the output as template expansion result.

    Results are cached / reused within the PreProcessor.

    :param tokenizer: Optionally, you can provide a non-standard tokenizer,
        e.g., one with a customized syntax.
    """

    def __init__(self, tokenizer: Optional[Tokenizer] = None) -> None:
        self._tokenizer = Tokenizer() if tokenizer is None else tokenizer
        # Tokenizer to be used by the PreProcessor. Currently, the same is
        # used for all files and for recursive evaluations the might start.
        # But this might change in the future.
        self._global_evaluation_context = GlobalEvaluationContext()
        # Context information for all template expansions happening under the
        # current PreProcessor.

    @staticmethod
    def diff(str1: str, str2: str, fromfile_txt: str, tofile_txt: str) -> str:
        """Compare two multi-line strings. If they are equal, return the empty string.
        Otherwise, return the result of the comparison in human-readable form.
        Currently, function *difflib.context_diff* is used for the comparison."""
        if str1 == str2:
            return ""
        return "".join(
            difflib.context_diff(
                str1.splitlines(keepends=True),
                str2.splitlines(keepends=True),
                fromfile_txt,
                tofile_txt,
            )
        )

    def template_script(
        self, template_file: FileName, trace_parsing: bool = False
    ) -> str:
        """
        Generate and return a template script for the template file.
        The method is typically only used for debugging purposes.

        :param template_file: Template to expand.
        :param trace_parsing: Print parsing log to stderr.
        """
        template = read_file(template_file)
        template_script = TemplateScript(
            str(template_file), template, self._tokenizer, trace_parsing
        )
        return str(template_script)

    def expand_file(
        self,
        template_file: FileName,
        diffs_to_template: bool = False,
        trace_parsing: bool = False,
        trace_evaluation: bool = False,
    ) -> str:
        """
        Load a template, expand it (recursive expansions may happen), and return the
        result.

        :param template_file: Template to expand.
        :param diffs_to_template: If True, the differences between the original
           template and the output generated by the macro expansion
           are returned, and the empty string if there are no differences.
        :param trace_parsing: Print parsing log to stderr.
        :param trace_evaluation: Print evaluation log to stderr.
        """
        template = read_file(template_file)
        template_script = TemplateScript(
            str(template_file),
            template,
            self._tokenizer,
            trace_parsing,
            trace_evaluation,
        )

        try:
            result = evaluate_template_script(
                template_script=template_script,
                tokenizer=self._tokenizer,
                global_evaluation_context=self._global_evaluation_context,
                already_imported_files=set[str](),
                globals_dict=None,
            )
        except Exception as exc:
            note = (
                f"Error occurred when expanding the following template: "
                f"  {str(template_file)}"
            )
            # Depending on the used Python version, one of the following will happen.
            if hasattr(exc, "add_note"):  # pragma: no cover
                exc.add_note(note)
                raise
            raise RuntimeError(note) from exc  # pragma: no cover

        return (
            self.diff(template, result, "template", "expansion result")
            if diffs_to_template
            else result
        )

    def expand_file_to_file(
        self,
        template_file: FileName,
        result_file: FileName,
        trace_parsing: bool = False,
        trace_evaluation: bool = False,
        diffs_to_result_file: bool = False,
    ) -> str:
        """Load a template file, expand it, and save the result to a file.

        :param template_file: Template to expand.
        :param result_file: File to store the results in.
        :param diffs_to_result_file: If True, nothing is written to the result_file.
           Instead, the differences between the content of the result_file
           and the output generated by the macro expansion are returned,
           or the empty string, if there are no differences.
        :param trace_parsing: Print parsing log to stderr.
        :param trace_evaluation: Print evaluation log to stderr.
        """
        result = self.expand_file(
            template_file,
            trace_parsing=trace_parsing,
            trace_evaluation=trace_evaluation,
        )
        if diffs_to_result_file:
            content = read_file(result_file)
            return self.diff(content, result, "current content", "expansion result")
        else:
            write_file(result_file, result)
            return ""
